# Accuracy Assessment Python v. 1.4

#  Created by:  Nathan Jennings
#               nate@jenningsplanet.com
#               www.jenningsplanet.com

# Created on:  03.26.2009
# Updated on:  03.01.2011

'''

Computes an Error Matrix for use with land cover classification projects.

The program creates a custom built table based on the number of unique
land cover (information) classes from a table containing known accuracy assessment polygon
values representing unique classfied (categorized) land cover types and the
majority land cover type computed from the ZonalStatisticsAsTable function
within the Zonal ArcToolbox within the Spatial Analyst Toolbox.  Majority is
one of the statistics computed from this tool.

All of the tools use the ArcToolbox alias name so there is no need to write specific
lines of code to the respective toolboxes.


REQUIREMENTS:

1.  This program requires the Spatial Analyst Extension for ArcGIS.  The
    ZonalStatistics as Table tool is used in this script and requires the
    Spatial Analyst extension.
    
2.  Currently, the arcgisscripting object uses 9.3, so while and for loops
    may need to be reconstructed if using 9.2.  The script has been tested with
    ArcGIS 9.3 and ArcGIS 10 on Windows XP.  Other permissions may need to be set
    for Windows 7.

INPUTS:



1.  Accuracy Assessment Polygon file.  The file must contain a unique ID
    that represents the specific AA site.  The file must also contain a field
    that contains that actual land cover value (information class), represented
    by an integer value (Parameter 3).

    For example, Water = 1, Grass = 2, Shrub = 3, etc.

2.  Unique ID field for the AA file to identify each unique accuracy assessment site
    (used for the Zonal Statistics function and the Table Join (see code)).

3.  Land Cover (Information) Class Field. Must be an integer value.

4.  Classified Image file.  This file contains "thematic values" that represent
    the specific land cover types (i.e. information classes).  These numbers
    correspond to the same values as the accuracy assessment file.

    Currently as of 03.01.2011, there is no check to make sure the image is a thematic type.

5.  The total number of unique land cover (information) classes in the classified image.
    For example, if there are 11 unique land cover types, the this number
    will be 11.  Currently, this value must be provided by the user of this
    tool.

6.  Folder where the Error Matrix table will be written.  This will be a folder.
    ***Currently no support to create a pGDB or fGDB table.

7.  Error Matrix table name (the output table).  If a .dbf extension is not provided
    the ".dbf" will be added by the script.


FUTURE ADDITIONS:

Check for thematic image type.

An output to CSV format will be added.

Review User's/Producer's Confidence Interval Calculations

_______________________________________________


UPDATES:

03.01.2011

Cleaned up User's, Producer's, and CIs statistics.  Some statistics need to be computed
by the total number of land cover classes, not number of rows, or number of rows - 1.
Number of Land Cover Classes = number of rows - 2 in the calculations.

Added a variable for maj_field so that the table join is more flexible.
maj_field is the "second" real column in the majority table that is created as part of
this process.

Added a check to see if the Spatial Analyst license is available for check out.  This means
that the user must have Spatial Analyst, otherwise an error message is raised.  If Spatial
Analyst exists and the user has not turned it on, the code "checks out" a license to use
to process the data in this script.

10.10.2010

Changed order of input parameters.
Added parameter to solicit land cover field


03.28.2010

Added Zonal Statistics to Table function.
Script uses only the accuracy assessement polygon file and classified image
as inputs.

Added Tool GUI and parameters to run within ArcToolbox.

03.27.2010

Completed all statistics.  Need to validate Confidence Interval values.


02.09.2010

Completed DRAFT version of all statistics (except overall and KHAT).

03.26.2009

Original draft of the program that included the Error Matrix format, and
the population of the Error Matrix.

Based on Accuracy Assessment Program written in VBA for ArcGIS.

'''


# Import system modules
import sys, string, os, arcgisscripting, traceback, time, datetime
from decimal import *

getcontext().prec = 2

# traceback, time, and datetime python modules are used for error tracking and time logging functions

# Create the Geoprocessor object
gp = arcgisscripting.create(9.3)   # this syntax will work for ArcGIS 10

gp.Overwriteoutput = 1

# Check out any necessary licenses  (REQUIRED for this tool to run)
try: 
    if gp.CheckExtension("spatial") == 'Available':
        gp.CheckOutExtension("spatial")
    else:

        raise LicenseError

except LicenseError:   # error handler for the Spatial Analyst check out license.

    print 'Spatial Analyst license is unavailable \n'
    print 'Make sure to check Spatial Analyst license or license manager'



# !!!! Spatial Analyst is required to run this program !!!!


#gp.workspace = "C:\\GIS Programming\\Geog_376\\Spring2009\\Instructor_Code"

#gp.workspace = "e:\\gis_programming\\accuracy_assessment"   # The workspace for the data

gp.workspace = sys.argv[6]   # This path will be used to store the Error Matrix table
#check_ws = gp.workspace
#gp.AddMessage("Workspace is: " + check_ws)


#if check_ws[:4] == ".mdb" or check_ws[:4] == ".fgd":
#    raise "BadWorkspace"

#gp.workspace = "c:\\gis programming\\application_development\\accuracy_assessment"
'''
if check_ws[:4] == ".mdb" or check_ws[:4] == ".fgd":
    gp.AddMessage("pGDB or fGDB")
    table_outpath = check_ws
    maj_table = "maj_table"
    EM_table = sys.argv[6]
    if EM_table[:4] == ".dbf":
        EM_table[:4]
    zone_root = maj_table

else:
'''    
table_outpath = gp.workspace + os.sep    # variable used to "create" the Error Matrix table
                                # a parameter for the "CreateTable" function below

maj_table = "maj_table.dbf"  # majority table, this is an intermediate for this process

EM_table = sys.argv[7]  # name of the Erorr matrix table (DBF format, currently supported)
                        # if the .dbf extension is not added, one will be added
if EM_table[:4] <> ".dbf":
    EM_table = EM_table + ".dbf"

zone_root = maj_table[:-4]
    

gp.AddMessage("Table Outpath: " + table_outpath)

# parameters for zonal majority table

#AA_fc = "aa_polys.shp" (or AA_Sites) # AA polygon file

AA_fc = sys.argv[1]   # The accuracy assessment polygon file


AA_fl = "aa_poly_fl"    # AA Feature Layer name used in the Addjoin below
#AA_root = AA_fc[:-4]    # Root name used in the Search Cursor below

basefname = os.path.basename(AA_fc)
splitfname = os.path.splitext(basefname)
AA_root = splitfname[0]

gp.AddMessage("AA Root: " + AA_root)

#zone_field = "AA_ID"       # Unique ID field for the specific AA sites

zone_field = sys.argv[2]    # These are the specific accuracy assessment polygon numbers or IDs

# Image file

#img_file = "super_recode.img"       # Classified (thematic data type) image file name

img_file = sys.argv[4]

#maj_table = "maj_table.dbf"         # Majority Table name (a temp file that
                                    # does not need to be saved)



#Num_LC = 11                 # The number of unique land cover classes (Information Classes)
Num_LC  = int(sys.argv[5])   # in the classified image (even if they not present
                            # in the classified image)

gp.AddMessage("Number of Unique Land Cover Classes: " + str(Num_LC))
# Validate the number of unique land cover classes

# Table and field name inputs
#inTable = sys.argv[1]
#inField = sys.argv[2]




# Error Matrix Table

#EM_table = "EM_table_test.dbf"      # Error Matrix Table (the output file)
#EM_table = sys.argv[6]

#if EM_table[:4] <> ".dbf" and :
#    EM_table = EM_table + ".dbf"

gp.AddMessage("EM Table: " + EM_table)

# Reference File name

#ref_file = "ref_out.dbf"
#ref_TV = "ref_TV"
#ref_root = ref_file[:-4]
#print ref_root

# Zone file name

#zone_file = "zone_out.dbf"

#zone_root = zone_file[:-4]
#zone_root = maj_table[:-4]      # Root name used in the Search Cursor below
zone_TV = "zone_TV"             # Table View name used in the Add Join below

#print zone_root

# Enter number of land cover (information) classes to compute Error Matrix for
gp.AddMessage("Getting Unique LC values...")
#lc_field = "CL_CODE"
lc_field = sys.argv[3]   # this is the field that contains the specific classified values
                         # i.e. information class values; a number that represents the specific
                         # recoded/reclassified values
lc_count = 0
# Search the AA Feature Class
AA_rows = gp.SearchCursor(AA_fc)

AA_row = AA_rows.Next()
gp.AddMessage("Searching...")
# Create an empty list
uniqueList = []
while AA_row:
    # If the value is not already in the list, append it
    if AA_row.GetValue(lc_field) not in uniqueList:
        uniqueList.append(AA_row.GetValue(lc_field))
        lc_count = lc_count + 1
    AA_row = AA_rows.Next()
# Sort the list alphanumerically    
uniqueList.sort()

# A check
gp.AddMessage("Number of Unique Land Cover Types: " + str(lc_count))


# Start the Error Matrix process

try:
    print "Computing Error Matrix"
    gp.AddMessage("Computing Error Matrix...")
    # Compute Zonal Majority Table from AA polygon file and classified image
    
    if gp.Exists(maj_table):
        gp.Delete_management(maj_table)

    print "Creating Zonal Statistics Table"
    gp.AddMessage("Creating Zonal Statistics Table...")
    
    gp.ZonalStatisticsAsTable_sa(AA_fc, zone_field, img_file, maj_table, "DATA")

    fieldList = gp.ListFields(maj_table)

    maj_field = fieldList[1].name   # This field is used in the table join below
                                    # and may be different depending on the version
                                    # of ArcGIS.
                                    # Always the first column after the OID, FID, or OBJECTID
    

    print "Finished Computing the Zonal Statistics Table"
    gp.AddMessage ("Finished Computing the Zonal Statistics Table")


    if gp.Exists(EM_table):
        gp.Delete_management(EM_table)

    gp.AddMessage("EM Table Path: " + table_outpath)      
    gp.CreateTable_management(table_outpath, EM_table)

    gp.AddMessage ("Created Table: " + table_outpath + EM_table)
   
    x=1
    #gp.AddMessage("X: " + str(x))

    # Add columns to EM table
    while (x < Num_LC + 1):
        field_name = "LC_" + str(x)
        gp.AddMessage("Adding Field Name: " + field_name + " to Error Matrix Table")
        gp.AddField_management(EM_table,field_name, "SHORT", "3")
        x = x + 1

    # Add Users, U CI High/Low, Producers, P CI High/Low, Overall, KHAT

    #gp.AddField_management(EM_table,"TOTAL" + str(x), "short", "3")
    gp.AddField_management(EM_table,"TOTAL", "DOUBLE", "6", "2")
    gp.AddField_management(EM_table,"USERS", "DOUBLE", "6", "2")
    gp.AddField_management(EM_table,"U_CI_LOW", "DOUBLE", "6", "2")
    gp.AddField_management(EM_table,"U_CI_HIGH", "DOUBLE", "6", "2")
    gp.AddField_management(EM_table,"PRODUCERS", "DOUBLE", "6", "2")
    gp.AddField_management(EM_table,"P_CI_LOW", "DOUBLE", "6", "2")
    gp.AddField_management(EM_table,"P_CI_HIGH", "DOUBLE", "6", "2")
    gp.AddField_management(EM_table,"OVERALL", "DOUBLE", "6", "2")
    gp.AddField_management(EM_table,"KHAT", "DOUBLE", "6", "2")




    print "Created the Output Error Matrix Table"
    gp.AddMessage("Created the Output Error Matrix Table")

    gp.DeleteField_management(EM_table, "Field1")  # This default field may show up
                                                   # when creating a .DBF file
                                                   # Currently, not checked for
                                                   # as of 03.01.2011
    
    # Add rows to EM table

    print "Inserting rows for each land cover (information) class..."
    gp.AddMessage("Inserting rows for each land cover (information) class...")

    rows = gp.InsertCursor(EM_table)
    
    x=0
    while ( x <=(Num_LC + 1)):
                
        row = rows.NewRow()
        rows.InsertRow(row)
        x=x+1

    #print "Added Rows"
    # Make TableViews of Ref and Zone files
    '''
    if gp.exists(ref_TV):
        gp.Delete(ref_TV)

    gp.MakeTableView_management(ref_file, ref_TV)
    '''

    if gp.exists(AA_fl):
        gp.Delete_management(AA_fl)

    gp.MakeFeatureLayer_management(AA_fc, AA_fl)
    
   
    if gp.exists(zone_TV):
        gp.Delete_management(zone_TV)
    
    #gp.MakeTableView_management(zone_file, zone_TV)

    gp.MakeTableView_management(maj_table, zone_TV)
    
    # Join Ref file to Majority file

    print "Joining Tables..."

    #gp.AddJoin_management(ref_TV, zone_field, zone_TV, "VALUE", "KEEP_ALL")
    gp.AddJoin_management(AA_fl, zone_field, zone_TV, maj_field, "KEEP_ALL") # "VALUE" "AA_ID"

    #fieldList = gp.ListFields(ref_TV)
    fieldList = gp.ListFields(AA_fl)

    #for field in fieldList:       # This structure will need to be revised
    #    print field.Name          # for pre 9.3 arcgisscripting objects


    #print "Joined table"

    # Read data from the Joined Table

    #rows = gp.SearchCursor(ref_TV)
    rows = gp.SearchCursor(AA_fl)
    row = rows.next()

    print "Computing Error Matrix Statistics..."
    gp.AddMessage("Computing Error Matrix Statistics...")
    
    count = 1
    while row:
        #print "Search Cursor Counter: " + str(count)
        #ref_val = row.GetValue(ref_root + ".CL_CODE")
        ref_val = row.GetValue(AA_root + "." + lc_field)
        maj_val = row.GetValue(zone_root + ".Majority")

        #print "Ref val: %i     Maj val: %i" % (ref_val, maj_val)

        # Get the proper row and column in the EM table and increment the proper value
        expression = "OID = " + str(maj_val)
        # print expression
        EM_rows = gp.UpdateCursor(EM_table, "OID = " + str(maj_val))
        #EM_count = gp.GetCount_management(EM_rows)
        #print EM_count
        EM_row = EM_rows.next()

        while EM_row:
            EM_val = EM_row.GetValue("LC_" + str(ref_val))
            # print "EM Val: " + str(EM_val)
                                     
            EM_row.SetValue("LC_" + str(ref_val), EM_val + 1)
            EM_rows.UpdateRow(EM_row)
            EM_row = EM_rows.next()
        count = count + 1
        row = rows.next()

    #####    Check this...  print "Finished incrementing rows"

    #print "Num Rows: " + str(count)

   

    # Calculate Row and Column values

    # Get an EM Row

    
    #EM_rows = gp.UpdateCursor(EM_table, "OID = " + str(maj_val))
    EM_rows = gp.UpdateCursor(EM_table)    # Get all rows
   
    EM_row = EM_rows.next()    # Skip the first row
    EM_row = EM_rows.next()    # Actual Rows start here
    result = gp.GetCount_management(EM_table)
    num_rows = int(result.GetOutput(0))

    cnt = 1

    total = 0
    overall_total = 0
    
    #print "Num Rows: " + str(num_rows)
    while EM_row:
        
        col = 1
        row_total = 0           # Reset col index to 0
        

        while col < (num_rows - 1):    # Loop over all columns for the row
                                 # to do the row sum

            # Comput Row total for User's AA
            #print EM_row.GetValue("LC_" + str(col))
            row_total = row_total + EM_row.GetValue("LC_" + str(col))
            col = col + 1

        
        prod_rows = gp.UpdateCursor(EM_table)   # Update cursor to read values from EM table
                                                # for Producer's accuracy
                                                
        prod_row = prod_rows.next()  # Skip blank row
        prod_row = prod_rows.next()  # First real row

        col_index = 1
        col_total = 0
        row_count = 0

        #gp.AddMessage("Num_rows: " + str(num_rows))

        while prod_row and cnt <= (num_rows - 1):
        
            

            #print "record num: " + str(cnt)
            
            #Loop over rows except the last one
            # Comput column totals for Producer's AA
            
            if cnt < (num_rows - 1):
                col_total = col_total + prod_row.GetValue("LC_" + str(cnt))

            row_count = row_count + 1

            if row_count == (num_rows - 1):     # if the row is the last row in the table
                if cnt < (num_rows - 1):
                    prod_row.SetValue("LC_" + str(cnt), col_total)    # Set the last row and column value to the column total
                    prod_rows.UpdateRow(prod_row)     # Update the value (i.e. producers column total)
                col_index = col_index + 1
                col_total = 0
            prod_row = prod_rows.next()

        # Compute row sum and User's Accuracy

        #print "Row Cnt before correct_val: " + str(cnt)
        if cnt < (num_rows - 1):
            correct_val = EM_row.GetValue("LC_" +  str(cnt))
        users_aa = correct_val / (row_total + 0.00000001) * 100   # computes User's AA
        row_avg = row_total / ((num_rows-2) + 0.00000001)    # Computes row average for variance calculation

        # Compute Row Variance
        col = 1
        row_var = 0
        while col < (num_rows - 1):      # sum each difference for each element in a single row
            row_var = row_var + (EM_row.GetValue("LC_" + str(col)) - row_avg)**2
            col = col + 1
            
        row_var = row_var /((num_rows - 2) + 0.00000001)   # take number from above and divide
                                                            # by the number of land cover classes

        # Compute Row Std Deviation

        row_sigma = row_var**0.5     # Take the square root

        #  Users Confidence Intervals
        rowLowCI = users_aa - (1.96 * (row_sigma / ((num_rows - 2)**0.5)))
        rowHighCI = users_aa + (1.96 * (row_sigma / ((num_rows - 2)**0.5)))

        EM_row.SetValue("TOTAL", row_total)
        
    
        if cnt < (num_rows - 1):
            EM_row.SetValue("USERS", users_aa)
            EM_row.SetValue("U_CI_LOW", rowLowCI)
            EM_row.SetValue("U_CI_HIGH", rowHighCI)

        # Work on Producer's Accuracy values
        
        prod_rows = gp.SearchCursor(EM_table, "OID = " + str(num_rows-1))# Get the last row of the EM table
                                                                         # this stores the producer's total (col total)
                                  
        result = gp.GetCount_management(EM_table)
        #print "Number of Search Rows: " + str(int(result.GetOutput(0)))
        prod_row = prod_rows.next()                          
        col = 1

        
        
        while prod_row:

            # Producer's info

            #print "CNT: " + str(cnt)
            #print "total: " + str(total)
            
            if cnt < (num_rows - 1): 
                # for the last row of the table that holds the column total
                prod_aa = correct_val / (prod_row.GetValue("LC_" + str(cnt)) + 0.00000001) * 100
                col_avg = prod_row.GetValue("LC_" + str(cnt)) / ((num_rows-2) + 0.00000001)

                #print "Total in prod_row: " + str(prod_row.GetValue("LC_" + str(cnt)))
                total = total + prod_row.GetValue("LC_" + str(cnt))

                
                
                #print "Total in prod_row: " + str(total) + " Prod value: " + str(prod_row.GetValue("LC_" + str(cnt)))
                #print "Prod AA: " + str(prod_aa)
                overall_total = overall_total + correct_val

                #print "Overall: " + str(overall_total)
            
            prod_row = prod_rows.next()    # No more rows in this case

        # Compute Producer's Variance

        v_rows = gp.SearchCursor(EM_table, "OID > 1" )  # Get the rows of the EM table
                                                        # a different Cursor is used here
                                                        # to differentiate from other cursors
        v_row = v_rows.next()

        #col_var = sum of (correct_val - col_avg ** 2

        #col_var = col_var + 

        while v_row:

            if cnt < (num_rows - 1):      # Don't include the last row (i.e. column totals)
                
                col = 1
                col_var = 0
                while col < (num_rows - 1):   # sum correct - avg squared; loop over rows for a single column
                                              # this is the numerator for the sum squares of the sigma calculation
                    col_var = col_var + (v_row.GetValue("LC_" + str(col)) - col_avg)**2
                    col = col + 1
                                
            v_row = v_rows.next()
        # The variance    
        col_var = col_var / ((num_rows - 2) + 0.00000001)   # divide the number from above by number of rows (i.e. observations)

        # Compute col Std Deviation

        col_sigma = col_var**0.5        # Take the square root


        # compute confidence intervals for producers AA
        colLowCI = prod_aa - (1.96 * (col_sigma / ((num_rows - 2)**0.5)))
        colHighCI = prod_aa + (1.96 * (col_sigma / ((num_rows - 2)**0.5)))

        
        if cnt < (num_rows - 1):
            EM_row.SetValue("PRODUCERS", prod_aa)
            EM_row.SetValue("P_CI_LOW", colLowCI)
            EM_row.SetValue("P_CI_HIGH", colHighCI)
        

        EM_rows.UpdateRow(EM_row)

        cnt = cnt + 1   # increment counter for next LC column to compute user total
        
        EM_row = EM_rows.next()

    #print "Total: " + str(total)

    EM_rows = gp.SearchCursor(EM_table, "OID > 0")    # Get all rows
   
    EM_row = EM_rows.next()    # Skip the first row

    cnt = 1
    diagsum = 0
    rowcolsum = 0
    
    while cnt < (num_rows - 1):

        diagsum = diagsum + EM_row.GetValue("LC_" + str(cnt))
        row_total = EM_row.GetValue("TOTAL")

        col_rows = gp.SearchCursor(EM_table, "OID = " + str(num_rows-1))   # Gets last row of EM Table
        col_row = col_rows.next()

        while col_row:
            col_total = col_row.GetValue("LC_" + str(cnt))   # Get the col total (Producer's total)
            col_row = col_rows.next()                        

        rowcolsum = rowcolsum + (row_total * col_total)

        cnt = cnt + 1
        EM_row = EM_rows.next()

    #print "Diag Sum: " + str(diagsum)
    #print "Row Total: " + str(row_total)
    #print "Col Total: " + str(col_total)
    #print "RowColSum: " + str(rowcolsum)
    #print "OverAll Total: " + str(overall_total)
        
    # Update the Overall and KHAT values in the EM Table

    overall_aa = ((overall_total) / Decimal(total)) * 100

    KHAT = (total * overall_total - rowcolsum) / (total**2 - rowcolsum)

    #print "OverAll: " + str(overall_aa)
    #print "KHAT: " + str(KHAT)
    
    OA_rows = gp.UpdateCursor(EM_table, "OID = 1")    # Get the first realrow


    
    
   
    OA_row = OA_rows.next()    # Get the row
    #OA_row = OA_rows.next()    # Actual Rows start here



    while OA_row:

        #print "Inside OA_row"
        tval = OA_row.GetValue("TOTAL")
        #print "T val: " + str(tval)
        OA_row.SetValue("OVERALL", float(overall_aa))
        OA_row.SetValue("KHAT", KHAT)
        OA_rows.UpdateRow(OA_row)
        OA_row = OA_rows.next()    


    
    # delete cursors to free up space and remove locks from EM table

    del row
    del rows
    del EM_row, OA_row, OA_rows, AA_row, AA_rows
    del EM_rows
    del prod_row
    del prod_rows
    del v_row
    del v_rows

      
    print "\nCompleted computing the Error Matrix Table: " + EM_table + \
          "\nMake sure to close the Python Shell and script file before " \
          "using the Error Matrix table."

    gp.AddMessage("\nCompleted computing the Error Matrix Table: " + EM_table + \
                  "\nMake sure to close the Python Shell and script file before " \
                  "using the Error Matrix table.")
    

    gp.CheckInExtension("spatial") 
    
    # Not presented here

    # Export table to CSV format   (See export_2_CSV.py) script
    


#except "BadWorkspace":
#    print "The Error Matrix Tool cannot use a pGDB or fGDB.  Choose a folder the workspace."
#    gp.AddError ("The Error Matrix Tool cannot use a pGDB or fGDB.  Choose a folder the workspace.")
    

except:

    tb = sys.exc_info()[2]
    tbinfo = traceback.format_tb(tb)[0]
    pymsg = "PYTHON ERRORS:\nTraceback Info:\n" + tbinfo + "\nError Info:\n     " +        str(sys.exc_type) + ": " + str(sys.exc_value) + "\n"
    msgs = "GP ERRORS:\n" + gp.GetMessages(2) + "\n"

    gp.AddError(msgs)
    gp.AddError(pymsg)

    print msgs
    print pymsg
    
    gp.AddMessage(gp.GetMessages(1))
    print gp.GetMessages(1)


